from collections.abc import Callable

from pyexasol import (
    ExaCommunicationError,
    ExaConnection,
    ExaRuntimeError,
    ExaStatement,
)


class DbConnection:
    """
    This is a pyexasol connection wrapper. It keeps the connection and requests a new
    one when the existing connection gets broken.

    Args:
        connection_factory:
            Supplied factory that creates a connection. The connection should be created
            with `fetch_dict`=True. The wrapper sets this option to True anyway. The
            dictionary option is required in order to present the result in a json form.
            This is what FastMCP expects from a tool.
    """

    def __init__(
        self, connection_factory: Callable[[], ExaConnection], num_retries: int = 2
    ) -> None:
        self._conn_factory = connection_factory
        self._num_retries = num_retries
        self._connection: ExaConnection | None = None

    def execute_query(self, query: str, snapshot: bool = True) -> ExaStatement:
        """
        Will make the set number of attempts to open or re-open the connection and
        execute the provided query. A repeated attempt may follow a CommunicationError
        or ExaRuntimeError. All other errors are considered unrecoverable and therefore
        will be propagated to the caller.

        If snapshot is True, which should be the mode of choice for querying metadata,
        the `meta.execute_snapshot` method will be called. Otherwise, it will use the
        normal `execute` method.
        """
        attempt = 1
        while True:
            if self._connection is None or self._connection.is_closed:
                self._connection = self._conn_factory()
                self._connection.options["fetch_dict"] = True
            try:
                if snapshot:
                    return self._connection.meta.execute_snapshot(query=query)
                return self._connection.execute(query=query)

            except (ExaCommunicationError, ExaRuntimeError):
                if not self._connection.is_closed:
                    self._connection.close()
                if attempt == self._num_retries:
                    raise
                attempt += 1



import re
from collections.abc import (
    Generator,
    Iterable,
)
from functools import cache
from typing import Any

import numpy as np
import stopwords
from rank_bm25 import BM25Plus


@cache
def _get_word_split_pattern() -> re.Pattern:
    r"""
    The pattern splits camel case and snake case words into multiple words. However,
    the camel case splitting only works for ASCII characters. More precisely, it will
    not work if a word starts or ends with a diacritic or non-latin character.
    The issue can be fixed using the regex library and the pattern
    (?<=\p{Ll})(?=\p{Lu})|_+
    """
    pattern = r"(?<=[a-z])(?=[A-Z])|_+"
    return re.compile(pattern, flags=re.MULTILINE)


@cache
def _get_word_extract_pattern() -> re.Pattern:
    pattern = r"\w+"
    return re.compile(pattern, flags=re.MULTILINE | re.UNICODE | re.IGNORECASE)


def _extract_raw_words(sentences: Iterable[str]) -> Generator[None, None, str]:
    for sentence in sentences:
        for match in _get_word_extract_pattern().finditer(
            _get_word_split_pattern().sub(" ", sentence)
        ):
            yield match.group().lower()


def extract_words(sentences: Iterable[str], language: str | None = None) -> list[str]:
    """
    Extracts words from all texts in the provided collection.
    All words in camel case or snake case get split into multiple words.
    All extracted words are returned lowercased.
    If the language is specified, a stopwords filter is applied.
    """
    words = list(_extract_raw_words(sentences))
    if language:
        # An error will be raised if the language is misspelled or not supported.
        return stopwords.clean(words, language.lower(), safe=False)
    return words


def get_match_scores(corpus: list[list[str]], keywords: list[str]) -> list[float]:
    """
    Assigns a keyword matching score for each text in the corpus .
    """
    bm25 = BM25Plus(corpus)
    return bm25.get_scores(keywords).tolist()


def _clipped_k_means(points: np.ndarray, max_iters=100) -> np.ndarray:
    """
    A simple implementation of K-means clustering for the special case:
        * 1-dimensional points,
        * 2 clusters,
        * the first centroid is clipped at the highest point value.
    The algorithm aims at picking one or a few outliers at the top end of the point
    value spectrum. The normal k-means may not work well in this case.
    """
    centroids = np.expand_dims(np.array([points.max(), points.mean()]), 1)
    prev_centroid = centroids[1, 0]
    labels = np.zeros_like(points, dtype=int)

    for _ in range(max_iters):
        # If both centroids collapsed into one point, stop
        if np.isclose(centroids[0, 0], centroids[1, 0]):
            break

        # Normal assignment step: assign each point to the closest centroid
        distances = np.square(points - centroids)
        labels = np.argmin(distances, axis=0)

        # If all points converged to cluster 0 (very unlikely case), then stop.
        if np.sum(labels) == 0:
            break

        # Update step, only update the centroid 1
        centroids[1, 0] = points[labels == 1].mean()

        # Check for convergence
        if np.isclose(centroids[1, 0], prev_centroid):
            break
        prev_centroid = centroids[1, 0]
    return labels


def top_score_indices(scores: list[float]) -> list[int]:
    """
    Divides the scores in the list into two clusters. Returns indices of scores in
    the cluster with higher scores. The returned indices are sorted by the score
    in the descending order.
    """
    labels = _clipped_k_means(np.asarray(scores))
    res = np.where(labels == 0)[0].tolist()
    return sorted(res, key=lambda i: scores[i], reverse=True)


def keyword_filter(
    input_rows: list[dict[str, Any]],
    key_phrases: list[str],
    language: str | None = None,
) -> list[dict[str, Any]]:
    """
    For each row in the input list, computes the keyword matching score. Returns the
    rows ordered by the score in descending order. Filters out rows with relatively
    low scores.

    Args:
        input_rows:
            list of input rows, formatted as {column_name: column_value} dictionaries.
        key_phrases:
            list of key_phrases.
        language:
            The language, e.g. german, the texts in the `input_rows` and the key phrases
            are written in.
    """
    keywords = extract_words(key_phrases, language)
    corpus = [
        extract_words(filter(lambda v: isinstance(v, str), di.values()), language)
        for di in input_rows
    ]
    scores = get_match_scores(corpus, keywords)
    indices = top_score_indices(scores)
    return [input_rows[i] for i in indices]



import re


def _exa_type_pattern() -> str:
    # A number surrounded by spaces
    n = r"\s*\d+\s*"
    # An optional number in brackets
    _1n_ = rf"(?:\s*\({n}\))?"
    # An optional two numbers in brackets
    _2n_ = rf"(?:\s*\({n}(?:,{n})?\))?"
    # An optional character set
    _char_set = rf"(?:(?:\s+CHARACTER\s+SET)?\s+(?:ASCII|UTF8))?"
    # An optional number of characters followed by an optional character set.
    _char_ = rf"(?:\s*\({n}(?:CHAR\s*)?\))?{_char_set}"

    exa_type_list = [
        "BOOL(?:EAN)?",
        f"DEC(?:IMAL)?{_2n_}",
        "INT(?:EGER)?",
        "BIGINT",
        "SHORTINT",
        "SMALLINT",
        "TINYINT",
        f"NUMBER{_2n_}",
        f"NUMERIC{_2n_}",
        r"DOUBLE(?:\s+PRECISION)?",
        "FLOAT",
        "REAL",
        rf"CHAR(?:ACTER)?(?:\s+VARYING)?{_char_}",
        rf"VARCHAR2?{_char_}",
        rf"CHARACTER\s+LARGE\s+OBJECT{_char_}",
        f"CLOB{_char_}",
        rf"LONG\s+VARCHAR{_char_set}",
        f"NCHAR{_char_}",
        f"NVARCHAR2?{_char_}",
        "DATE",
        rf"TIMESTAMP{_1n_}(?:\s+(?:WITH\s+LOCAL|WITHOUT)\s+TIME\s+ZONE)?",
        rf"INTERVAL\s+DAY{_1n_}\s+TO\s+SECOND{_1n_}",
        rf"INTERVAL\s+YEAR{_1n_}\s+TO\s+MONTH",
        rf"GEOMETRY{_1n_}",
        rf"HASHTYPE(?:\s*\(\s*\d+\s+(?:BIT|BYTE)\s*\))?",
    ]
    type_choice = "|".join(exa_type_list)
    return f"(?:{type_choice})"


exa_type_pattern = _exa_type_pattern()
"""
The exact RE pattern of an Exasol SQL type.
"""

identifier_pattern = r"[^\W\d]\w*"
"""
The RE pattern of an identifier, e.g. a parameter name.
"""

quoted_identifier_pattern = f'(?:{identifier_pattern}|"(?:[^"]|"")+")'
"""
The RE pattern of an identifier optionally enclosed in double quotes.
The double quotes may enclose any text. The double quote characters in it
are escaped with double double quotes.
"""

parameter_pattern = rf"{identifier_pattern}\s+{exa_type_pattern}"
"""
The RE pattern of a function parameter, consisting of the parameter name and SQL type.
"""

quoted_parameter_pattern = rf"{quoted_identifier_pattern}\s+{exa_type_pattern}"
"""
The RE pattern of a script parameter, consisting of the parameter name, optionally
enclosed in double quotes, and SQL type.
"""

regex_flags = re.IGNORECASE | re.MULTILINE | re.UNICODE
"""
The RE flags required to match the parameter names and SQL types.
"""

parameter_list_pattern = rf"(?:,?\s*{quoted_parameter_pattern}\s*(?=\)|,))*"
r"""
A parameter list matching pattern, matching zero, one or more parameters. In the text,
the list should be enclosed in parentheses. A parameter may follow a comma, if it's not
the first one: The lookahead symbol after the parameter should be either a comma or a
closing parenthesis: (?=\)|,).
"""



import re
from enum import Enum
from functools import cache

import sqlglot.expressions as exp

from exasol.ai.mcp.server.server_settings import (
    McpServerSettings,
    MetaListSettings,
)

INFO_COLUMN = "SUPPORT_INFO"
"""
Column with additional information, to be used for filtering.
"""

GROUP_CONCAT = "GROUP_CONCAT_"
"""
A surrogate name for the GROUP_CONCAT function, to be replaced with "GROUP_CONCAT"
in the syntax correction.
"""


class MetaType(Enum):
    SCHEMA = "SCHEMA"
    TABLE = "TABLE"
    VIEW = "VIEW"
    FUNCTION = "FUNCTION"
    SCRIPT = "SCRIPT"
    COLUMN = "COLUMN"


@cache
def _get_group_concat_pattern() -> re.Pattern:
    return re.compile(
        rf'{GROUP_CONCAT}\(\s*((?:DISTINCT)?\s+"\w+"),?(\s+ORDER\sBY\s+"\w+")?\s*\)'
    )


def _fix_group_concat(query_sql: str, separator: str | None = None) -> str:
    """
    Corrects the GROUP_CONCAT syntax, as the Exasol syntax is not yet supported.
    """
    pattern = _get_group_concat_pattern()
    repl_separator = f" SEPARATOR '{separator}'" if separator else ""
    return pattern.sub(rf"GROUP_CONCAT(\1\2{repl_separator})", query_sql)


def _get_meta_predicates(column: str, conf: MetaListSettings) -> list[exp.Predicate]:
    """
    Constructs predicates for the provided column using the LIKE and REGEXP_LIKE
    patterns in the provided configuration.
    """
    predicates: list[exp.Predicate] = []
    if conf.like_pattern:
        predicates.append(exp.column(column).like(conf.like_pattern))
    if conf.regexp_pattern:
        predicates.append(
            exp.func(
                "REGEXP_INSTR",
                exp.column(column),
                exp.Literal.string(conf.regexp_pattern),
            ).neq(0)
        )
    return predicates


def _inner_meta_query(
    meta_type: MetaType, output_types: list[MetaType], *predicates
) -> exp.Query:
    """
    Builds a query collecting names and comments for the given type of meta and putting
    them in a json string. The comment element is omitted if NULL.

    Args:
        meta_type:
            Metadata type to collect the data for.
        output_types:
            List of metadata types to go in the SELECT list. This should correspond to
            the `meta_type`. For example, the output type TABLE can be used only with
            the meta type COLUMN, but 'SCHEMA' can be used with any meta type.
        predicates:
            WHERE clause predicates.

    Example output for `meta_name` = 'TABLE', output_types = ['SCHEMA'].
    {
        'SCHEMA_NAME': '<schema name>',
        'OBJ_INFO': '{"TABLE": "<table1 name>", "COMMENT": "<table1 comment>"}'
    }
    """
    meta_name = meta_type.value
    select_list = [
        exp.column(f"{meta_name}_{out_type.value}").as_(out_type.value)
        for out_type in output_types
    ]
    query = (
        exp.Select()
        .select(
            *select_list,
            exp.func(
                "CONCAT",
                exp.Literal.string(f'{{"{meta_name}": "'),
                exp.column(f"{meta_name}_NAME"),
                exp.func(
                    "NVL2",
                    exp.column(f"{meta_name}_COMMENT"),
                    exp.func(
                        "CONCAT",
                        exp.Literal.string('", "COMMENT": "'),
                        exp.column(f"{meta_name}_COMMENT"),
                    ),
                    exp.Literal.string(""),
                ),
                exp.Literal.string('"}'),
            ).as_("OBJ_INFO"),
        )
        .from_(exp.Table(this=f"EXA_ALL_{meta_name}S", db="SYS"))
        .where(*predicates)
    )
    return query


@cache
def _info_concat_func() -> exp.Expression:
    return exp.func(
        "CONCAT",
        exp.Literal.string("["),
        exp.func(GROUP_CONCAT, exp.Distinct(expressions=[exp.column("OBJ_INFO")])),
        exp.Literal.string("]"),
    ).as_(INFO_COLUMN)


class ExasolMetaQuery:
    """
    A query builder class, constructing metadata queries based on the provided server
    configuration.
    """

    def __init__(self, config: McpServerSettings):
        self._config = config
        self._meta_conf = {
            MetaType.SCHEMA: config.schemas,
            MetaType.TABLE: config.tables,
            MetaType.VIEW: config.views,
            MetaType.FUNCTION: config.functions,
            MetaType.SCRIPT: config.scripts,
        }

    @property
    def config(self) -> McpServerSettings:
        return self._config

    def _get_column_eq_predicate(
        self, column: str, value: str, table: str | None = None
    ) -> exp.Predicate:
        """
        Builds an expression of a text column equality, considering the
        case sensitivity setting.
        """
        if self.config.case_sensitive:
            return exp.column(column, table=table).eq(exp.Literal.string(value))
        return exp.func("UPPER", exp.column(column, table=table)).eq(
            exp.Literal.string(value.upper())
        )

    def get_metadata(self, meta_type: MetaType, schema_name: str | None = None) -> str:
        """
        A generic metadata query. Collects the DB object name, schema and comment
        for a given object type. Applies visibility restrictions for the type of
        metadata in question and for the schema. Optionally, limits the output to
        the objects in one particular schema.

        Args:
            meta_type:
                Metadata type to collect the data for.
            schema_name:
                An optional schema name provided in the call to the tool. Ignored if
                meta_name=='SCHEMA'. In all other cases, if the name is specified, it
                will be included in the WHERE clause. Otherwise, the query will return
                objects from all visible schemas.
        """
        conf = self._meta_conf[meta_type]
        meta_name = meta_type.value
        meta_name_column = f"{meta_name}_NAME"
        select_list = [
            exp.column(meta_name_column).as_(conf.name_field),
            exp.column(f"{meta_name}_COMMENT").as_(conf.comment_field),
        ]
        predicates = _get_meta_predicates(meta_name_column, conf)
        if meta_type == MetaType.SCRIPT:
            predicates.append(exp.column("SCRIPT_TYPE").eq("UDF"))
        if meta_type != MetaType.SCHEMA:
            schema_column = f"{meta_name}_SCHEMA"
            if schema_name:
                predicates.append(
                    self._get_column_eq_predicate(schema_column, schema_name)
                )
            else:
                # Adds the schema restriction if specified in the settings.
                predicates.extend(
                    _get_meta_predicates(schema_column, self.config.schemas)
                )
            select_list.append(exp.column(schema_column).as_(conf.schema_field))
        query = (
            exp.Select()
            .select(*select_list)
            .from_(exp.Table(this=f"EXA_ALL_{meta_name}S", db="SYS"))
            .where(*predicates)
        )
        return query.sql(dialect="exasol", identify=True)

    def get_object_metadata(
        self, meta_type: MetaType, schema_name: str, obj_name: str
    ) -> str:
        """
        Builds a query requesting metadata for a given database object.
        """
        meta_name = meta_type.value
        query = (
            exp.Select()
            .select(exp.Star())
            .from_(exp.Table(this=f"EXA_ALL_{meta_name}S", db="SYS"))
            .where(
                exp.and_(
                    self._get_column_eq_predicate(f"{meta_name}_SCHEMA", schema_name),
                    self._get_column_eq_predicate(f"{meta_name}_NAME", obj_name),
                )
            )
        )
        return query.sql(dialect="exasol", identify=True)

    @cache
    def find_schemas(self) -> str:
        """
        Collects names, comments and the support information for schemas within
        the defined schema visibility.

        The support information includes the names and comments of the database
        objects within the schema. The visibility rules defined for each type of
        metadata is applied when the support information is collected.
        The support information is formated as json. For an example see the
        `_inner_meta_query`.
        """
        inner_queries = exp.Paren(
            this=exp.union(
                *(
                    _inner_meta_query(
                        meta_type,
                        [MetaType.SCHEMA],
                        predicate,
                        *_get_meta_predicates(
                            f"{meta_type.value}_NAME", self._meta_conf[meta_type]
                        ),
                    )
                    for meta_type, predicate in [
                        (MetaType.TABLE, ""),
                        (MetaType.VIEW, ""),
                        (MetaType.FUNCTION, ""),
                        (
                            MetaType.SCRIPT,
                            exp.column("SCRIPT_TYPE").eq(exp.Literal.string("UDF")),
                        ),
                    ]
                    if self._meta_conf[meta_type].enable
                ),
                dialect="exasol",
            )
        )

        predicates = _get_meta_predicates("SCHEMA_NAME", self._config.schemas)

        query = (
            exp.Select()
            .select(
                exp.column("SCHEMA_NAME", table="S").as_(
                    self._config.schemas.name_field
                ),
                exp.column("SCHEMA_COMMENT", table="S").as_(
                    self._config.schemas.comment_field
                ),
                exp.column(INFO_COLUMN, db="O"),
            )
            .from_(exp.Table(this="EXA_ALL_SCHEMAS", db="SYS").as_("S"))
            .where(*predicates)
            .join(
                exp.Select()
                .select(exp.column("SCHEMA"), _info_concat_func())
                .from_(inner_queries)
                .group_by("SCHEMA")
                .as_("O"),
                on=exp.column("SCHEMA_NAME", table="S").eq(
                    exp.column("SCHEMA", table="O")
                ),
            )
        )
        query_sql = query.sql(dialect="exasol", identify=True)
        return _fix_group_concat(query_sql, separator=", ")

    def find_tables(self, schema_name: str | None) -> str:
        """
        Collects names, comments and support information for tables and/or views,
        depending on what is enabled. The visibility rules defined for the schemas,
        tables and views are applied. An optional `schema_name`, if provided, restricts
        the listing of the tables and views to this particular schema.

        The support information includes the names and comments of the columns.
        It is formated as json. For an example see the `_inner_meta_query`.
        """
        if schema_name:
            predicates = [self._get_column_eq_predicate("COLUMN_SCHEMA", schema_name)]
        else:
            predicates = _get_meta_predicates("COLUMN_SCHEMA", self.config.schemas)
        inner_query = _inner_meta_query(
            MetaType.COLUMN, [MetaType.SCHEMA, MetaType.TABLE], *predicates
        )
        cte = (
            exp.Select()
            .select(exp.column("SCHEMA"), exp.column("TABLE"), _info_concat_func())
            .from_(exp.Paren(this=inner_query))
            .group_by("SCHEMA", "TABLE")
        )
        main_queries = [
            (
                exp.Select()
                .select(
                    exp.column(f"{meta_name}_NAME", table="T").as_(conf.name_field),
                    exp.column(f"{meta_name}_COMMENT", table="T").as_(
                        conf.comment_field
                    ),
                    exp.column(f"{meta_name}_SCHEMA", table="T").as_(conf.schema_field),
                    exp.column(INFO_COLUMN, table="C"),
                )
                .from_(exp.Table(this=f"EXA_ALL_{meta_name}S", db="SYS").as_("T"))
                .join(
                    "C",
                    on=exp.and_(
                        exp.column(f"{meta_name}_SCHEMA", table="T").eq(
                            exp.column("SCHEMA", table="C")
                        ),
                        exp.column(f"{meta_name}_NAME", table="T").eq(
                            exp.column("TABLE", table="C")
                        ),
                    ),
                )
                .where(*_get_meta_predicates(f"{meta_name}_NAME", conf))
            )
            for meta_name, conf in [
                ("TABLE", self._config.tables),
                ("VIEW", self._config.views),
            ]
            if conf.enable
        ]
        query = main_queries[0] if len(main_queries) == 1 else exp.union(*main_queries)
        query = query.with_(alias="C", as_=cte)
        query_sql = query.sql(dialect="exasol", identify=True)
        return _fix_group_concat(query_sql, separator=", ")

    def describe_columns(self, schema_name: str, table_name: str) -> str:
        """
        Gathers a list of columns in a given table.
        """
        conf = self._config.columns
        query = (
            exp.Select()
            .select(
                exp.column("COLUMN_NAME").as_(conf.name_field),
                exp.column("COLUMN_TYPE").as_(conf.type_field),
                exp.column("COLUMN_COMMENT").as_(conf.comment_field),
            )
            .from_(exp.Table(this="EXA_ALL_COLUMNS", db="SYS"))
            .where(
                exp.and_(
                    self._get_column_eq_predicate("COLUMN_SCHEMA", schema_name),
                    self._get_column_eq_predicate("COLUMN_TABLE", table_name),
                )
            )
        )
        return query.sql(dialect="exasol", identify=True)

    def describe_constraints(self, schema_name: str, table_name: str) -> str:
        """
        Gathers a list of constraints for a given table.
        """
        conf = self._config.columns
        query = (
            exp.Select()
            .select(
                exp.FirstValue(this=exp.column("CONSTRAINT_TYPE")).as_(
                    conf.constraint_type_field
                ),
                exp.case(exp.func("LEFT", exp.column("CONSTRAINT_NAME"), 4))
                .when(exp.Literal.string("SYS_"), exp.Null())
                .else_(exp.column("CONSTRAINT_NAME"))
                .as_(conf.constraint_name_field),
                exp.func(
                    GROUP_CONCAT,
                    exp.Distinct(expressions=[exp.column("COLUMN_NAME")]),
                    exp.Order(expressions=[exp.column("ORDINAL_POSITION")]),
                ).as_(conf.constraint_columns_field),
                exp.FirstValue(this=exp.column("REFERENCED_SCHEMA")).as_(
                    conf.referenced_schema_field
                ),
                exp.FirstValue(this=exp.column("REFERENCED_TABLE")).as_(
                    conf.referenced_table_field
                ),
                exp.func(
                    GROUP_CONCAT,
                    exp.Distinct(expressions=[exp.column("REFERENCED_COLUMN")]),
                    exp.Order(expressions=[exp.column("ORDINAL_POSITION")]),
                ).as_(conf.referenced_columns_field),
            )
            .from_(exp.Table(this="EXA_ALL_CONSTRAINT_COLUMNS", db="SYS"))
            .where(
                exp.and_(
                    self._get_column_eq_predicate("CONSTRAINT_SCHEMA", schema_name),
                    self._get_column_eq_predicate("CONSTRAINT_TABLE", table_name),
                )
            )
            .group_by("CONSTRAINT_NAME")
        )
        query_sql = query.sql(dialect="exasol", identify=True)
        return _fix_group_concat(query_sql)

    def get_table_comment(self, schema_name: str, table_name: str) -> str:
        """
        The query returns a single row with a comment for a given table or view.
        """
        # `table_name` can be the name of a table or a view.
        # This query tries both possibilities. The UNION clause collapses
        # the result into a single non-NULL distinct value.
        queries = [
            exp.Select()
            .select(exp.column(f"{meta_name}_COMMENT").as_("COMMENT"))
            .from_(exp.Table(this=f"EXA_ALL_{meta_name}S", db="SYS"))
            .where(
                exp.and_(
                    self._get_column_eq_predicate(f"{meta_name}_SCHEMA", schema_name),
                    self._get_column_eq_predicate(f"{meta_name}_NAME", table_name),
                )
            )
            for meta_name in ["TABLE", "VIEW"]
        ]
        query = exp.Union(this=queries[0], expression=queries[1]).limit(1)
        return query.sql(dialect="exasol", identify=True)



import re
from functools import cache
from typing import (
    Annotated,
    Any,
)

from fastmcp import FastMCP
from pydantic import Field
from sqlglot import (
    exp,
    parse_one,
)
from sqlglot.errors import ParseError

from exasol.ai.mcp.server.db_connection import DbConnection
from exasol.ai.mcp.server.keyword_search import keyword_filter
from exasol.ai.mcp.server.meta_query import (
    INFO_COLUMN,
    ExasolMetaQuery,
    MetaType,
)
from exasol.ai.mcp.server.parameter_parser import (
    FuncParameterParser,
    ScriptParameterParser,
)
from exasol.ai.mcp.server.parameter_pattern import (
    parameter_list_pattern,
    regex_flags,
)
from exasol.ai.mcp.server.server_settings import (
    ExaDbResult,
    McpServerSettings,
)

TABLE_USAGE = (
    "In an SQL query, the names of database objects, such as schemas, "
    "tables and columns should be enclosed in double quotes. "
    "A reference to a table should include a reference to its schema. "
    "The SELECT column list cannot have both the * and explicit column names."
)

SCHEMA_NAME_TYPE = Annotated[str, Field(description="name of the database schema")]

OPTIONAL_SCHEMA_NAME_TYPE = Annotated[
    str | None, Field(description="optional name of the database schema", default="")
]

KEYWORDS_TYPE = Annotated[list[str], "list of keywords to filter and order the result"]

TABLE_NAME_TYPE = Annotated[str, "name of the table"]

FUNCTION_NAME_TYPE = Annotated[str, "name of the function"]

SCRIPT_NAME_TYPE = Annotated[str, "name of the script"]


@cache
def _get_emits_pattern() -> re.Pattern:
    pattern = rf"EMITS\s*\({parameter_list_pattern}\)"
    return re.compile(pattern, flags=regex_flags)


def verify_query(query: str) -> bool:
    """
    Verifies that the query is a valid SELECT query.
    Declines any other types of statements including the SELECT INTO.
    """

    # Here is a fix for the SQLGlot deficiency in understanding the syntax of variadic
    # emit UDF. The EMITS clause in the SELECT statement is currently not recognised.
    # To let SQLGlot validate the query, this clause must be pinched away.
    query = _get_emits_pattern().sub("", query)

    try:
        ast = parse_one(query, read="exasol")
        if isinstance(ast, exp.Select):
            return "into" not in ast.args
        return False
    except ParseError:
        return False


def remove_info_column(result: list[dict[str, Any]]) -> list[dict[str, Any]]:
    """
    Removes the column with extra information, collected for data filtering.
    """
    for row in result:
        if INFO_COLUMN in row:
            row.pop(INFO_COLUMN)
    return result


class ExasolMCPServer(FastMCP):
    """
    An Exasol MCP server based on FastMCP.

    Args:
        connection:
            pyexasol connection wrapper.
        config:
            The server configuration.
    """

    def __init__(self, connection: DbConnection, config: McpServerSettings) -> None:
        super().__init__(name="exasol-mcp")
        self.connection = connection
        self.meta_query = ExasolMetaQuery(config)

    @property
    def config(self) -> McpServerSettings:
        return self.meta_query.config

    def _execute_meta_query(
        self, query: str, keywords: list[str] | None = None
    ) -> ExaDbResult:
        """
        Executes a metadata query and returns the result as a list of dictionaries.
        Applies the keyword fitter if provided.
        Removes the column with extra information that could be added to the result
        to assist filtering. This is necessary to avoid polluting the LLM's context
        with data it doesn't need at the current stage.
        """
        result = self.connection.execute_query(query).fetchall()
        if keywords:
            result = keyword_filter(result, keywords, language=self.config.language)
        return ExaDbResult(remove_info_column(result))

    def list_schemas(self) -> ExaDbResult:
        if not self.config.schemas.enable:
            raise RuntimeError("The schema listing is disabled.")

        query = self.meta_query.get_metadata(MetaType.SCHEMA)
        return self._execute_meta_query(query)

    def find_schemas(self, keywords: KEYWORDS_TYPE) -> ExaDbResult:
        if not self.config.schemas.enable:
            raise RuntimeError("The schema listing is disabled.")

        query = self.meta_query.find_schemas()
        return self._execute_meta_query(query, keywords)

    def list_tables(self, schema_name: SCHEMA_NAME_TYPE) -> ExaDbResult:
        table_conf = self.config.tables
        view_conf = self.config.views
        if (not table_conf.enable) and (not view_conf.enable):
            raise RuntimeError("Both the table and the view listings are disabled.")

        query = "\nUNION\n".join(
            self.meta_query.get_metadata(meta_type, schema_name)
            for meta_type, conf in zip(
                [MetaType.TABLE, MetaType.VIEW], [table_conf, view_conf]
            )
            if conf.enable
        )
        return self._execute_meta_query(query)

    def find_tables(
        self, keywords: KEYWORDS_TYPE, schema_name: OPTIONAL_SCHEMA_NAME_TYPE
    ) -> ExaDbResult:
        if (not self.config.tables.enable) and (not self.config.views.enable):
            raise RuntimeError("Both the table and the view listings are disabled.")

        query = self.meta_query.find_tables(schema_name)
        return self._execute_meta_query(query, keywords)

    def list_functions(self, schema_name: SCHEMA_NAME_TYPE) -> ExaDbResult:
        if not self.config.functions.enable:
            raise RuntimeError("The function listing is disabled.")

        query = self.meta_query.get_metadata(MetaType.FUNCTION, schema_name)
        return self._execute_meta_query(query)

    def find_functions(
        self, keywords: KEYWORDS_TYPE, schema_name: OPTIONAL_SCHEMA_NAME_TYPE
    ) -> ExaDbResult:
        if not self.config.functions.enable:
            raise RuntimeError("The function listing is disabled.")

        query = self.meta_query.get_metadata(MetaType.FUNCTION, schema_name)
        return self._execute_meta_query(query, keywords)

    def list_scripts(self, schema_name: SCHEMA_NAME_TYPE) -> ExaDbResult:
        if not self.config.scripts.enable:
            raise RuntimeError("The script listing is disabled.")

        query = self.meta_query.get_metadata(MetaType.SCRIPT, schema_name)
        return self._execute_meta_query(query)

    def find_scripts(
        self, keywords: KEYWORDS_TYPE, schema_name: OPTIONAL_SCHEMA_NAME_TYPE
    ) -> ExaDbResult:
        if not self.config.scripts.enable:
            raise RuntimeError("The script listing is disabled.")

        query = self.meta_query.get_metadata(MetaType.SCRIPT, schema_name)
        return self._execute_meta_query(query, keywords)

    def describe_columns(
        self, schema_name: SCHEMA_NAME_TYPE, table_name: TABLE_NAME_TYPE
    ) -> ExaDbResult:
        """
        Returns the list of columns in the given table. Currently, this is a part of
        the `describe_table` tool, but it can be used independently in the future.
        """
        if not self.config.columns.enable:
            raise RuntimeError("The column listing is disabled.")

        query = self.meta_query.describe_columns(schema_name, table_name)
        return self._execute_meta_query(query)

    def describe_constraints(
        self, schema_name: SCHEMA_NAME_TYPE, table_name: TABLE_NAME_TYPE
    ) -> ExaDbResult:
        """
        Returns the list of constraints in the given table. Currently, this is a part
        of the `describe_table` tool, but it can be used independently in the future.
        """
        if not self.config.columns.enable:
            raise RuntimeError("The constraint listing is disabled.")

        query = self.meta_query.describe_constraints(schema_name, table_name)
        return self._execute_meta_query(query)

    def get_table_comment(self, schema_name: str, table_name: str) -> str | None:
        query = self.meta_query.get_table_comment(schema_name, table_name)
        comment_row = self.connection.execute_query(query).fetchone()
        if comment_row is None:
            return None
        table_comment = next(iter(comment_row.values()))
        if table_comment is None:
            return None
        return str(table_comment)

    def describe_table(
        self, schema_name: SCHEMA_NAME_TYPE, table_name: TABLE_NAME_TYPE
    ) -> dict[str, Any]:

        conf = self.config.columns
        columns = self.describe_columns(schema_name, table_name)
        constraints = self.describe_constraints(schema_name, table_name)
        table_comment = self.get_table_comment(schema_name, table_name)

        return {
            conf.columns_field: columns.result,
            conf.constraints_field: constraints.result,
            conf.table_comment_field: table_comment,
            conf.usage_field: TABLE_USAGE,
        }

    def describe_function(
        self, schema_name: SCHEMA_NAME_TYPE, func_name: FUNCTION_NAME_TYPE
    ) -> dict[str, Any]:
        parser = FuncParameterParser(connection=self.connection, settings=self.config)
        return parser.describe(schema_name, func_name)

    def describe_script(
        self, schema_name: SCHEMA_NAME_TYPE, script_name: SCRIPT_NAME_TYPE
    ) -> dict[str, Any]:
        parser = ScriptParameterParser(connection=self.connection, settings=self.config)
        return parser.describe(schema_name, script_name)

    def execute_query(
        self, query: Annotated[str, Field(description="select query")]
    ) -> ExaDbResult:
        if not self.config.enable_read_query:
            raise RuntimeError("Query execution is disabled.")
        if verify_query(query):
            result = self.connection.execute_query(query, snapshot=False).fetchall()
            return ExaDbResult(result)
        raise ValueError("The query is invalid or not a SELECT statement.")



import json
import os
import re

import pyexasol
from pydantic import ValidationError
from pyexasol import ExaConnection

from exasol.ai.mcp.server.db_connection import DbConnection
from exasol.ai.mcp.server.mcp_server import ExasolMCPServer
from exasol.ai.mcp.server.server_settings import McpServerSettings

ENV_SETTINGS = "EXA_MCP_SETTINGS"
ENV_DSN = "EXA_DSN"
ENV_USER = "EXA_USER"
ENV_PASSWORD = "EXA_PASSWORD"


def _register_list_schemas(mcp_server: ExasolMCPServer) -> None:
    mcp_server.tool(
        mcp_server.list_schemas,
        description=(
            "The tool lists schemas in the Exasol Database. "
            "For each schema, it provides the name and an optional comment."
        ),
    )


def _register_find_schemas(mcp_server: ExasolMCPServer) -> None:
    mcp_server.tool(
        mcp_server.find_schemas,
        description=(
            "The tool finds schemas in the Exasol Database by looking for the "
            "specified keywords in their names and comments. The list of keywords "
            "should include common inflections of each keyword. "
            "For each schema it finds, it provides the name and an optional comment."
        ),
    )


def _register_list_tables(mcp_server: ExasolMCPServer) -> None:
    mcp_server.tool(
        mcp_server.list_tables,
        description=(
            "The tool lists tables and views in the specified schema of the "
            "the Exasol Database. For each table and view, it provides the "
            "name, the schema, and an optional comment."
        ),
    )


def _register_find_tables(mcp_server: ExasolMCPServer) -> None:
    mcp_server.tool(
        mcp_server.find_tables,
        description=(
            "The tool finds tables and views in the Exasol Database by looking "
            "for the specified keywords in their names and comments. The list of "
            "keywords should include common inflections of each keyword. "
            "For each table or view the tool finds, it provides the name, the schema, "
            "and an optional comment. An optional `schema_name` argument allows "
            "restricting the search to tables and views in the specified schema."
        ),
    )


def _register_list_functions(mcp_server: ExasolMCPServer) -> None:
    mcp_server.tool(
        mcp_server.list_functions,
        description=(
            "The tool lists functions in the specified schema of the Exasol "
            "Database. For each function, it provides the name, the schema, "
            "and an optional comment."
        ),
    )


def _register_find_functions(mcp_server: ExasolMCPServer) -> None:
    mcp_server.tool(
        mcp_server.find_functions,
        description=(
            "The tool finds functions in the Exasol Database by looking for "
            "the specified keywords in their names and comments. The list of "
            "keywords should include common inflections of each keyword. "
            "For each function the tool finds, it provides the name, the schema,"
            "and an optional comment. An optional `schema_name` argument allows "
            "restricting the search to functions in the specified schema."
        ),
    )


def _register_list_scripts(mcp_server: ExasolMCPServer) -> None:
    mcp_server.tool(
        mcp_server.list_scripts,
        description=(
            "The tool lists the user defined functions (UDF) in the specified "
            "schema of the Exasol Database. For each UDF, it provides the name, "
            "the schema, and an optional comment."
        ),
    )


def _register_find_scripts(mcp_server: ExasolMCPServer) -> None:
    mcp_server.tool(
        mcp_server.find_scripts,
        description=(
            "The tool finds the user defined functions (UDF) in the Exasol Database "
            "by looking for the specified keywords in their names and comments. The "
            "list of keywords should include common inflections of each keyword. "
            "For each UDF the tool finds, it provides the name, the schema, and an "
            "optional comment. An optional `schema_name` argument allows restricting "
            "the search to UDFs in the specified schema."
        ),
    )


def _register_describe_table(mcp_server: ExasolMCPServer) -> None:
    mcp_server.tool(
        mcp_server.describe_table,
        description=(
            "The tool describes the specified table or view in the specified "
            "schema of the Exasol Database. The description includes the list "
            "of columns and for a table also the list of constraints. For each "
            "column the tool provides the name, the SQL data type and an "
            "optional comment. For each constraint it provides its type, e.g. "
            "PRIMARY KEY, the list of columns the constraint is applied to and "
            "an optional name. For a FOREIGN KEY it also provides the referenced "
            "schema, table and a list of columns in the referenced table."
        ),
    )


def _register_describe_function(mcp_server: ExasolMCPServer) -> None:
    mcp_server.tool(
        mcp_server.describe_function,
        description=(
            "The tool describes the specified function in the specified schema "
            "of the Exasol Database. It provides the list of input parameters "
            "and the return SQL type. For each parameter it specifies the name "
            "and the SQL type."
        ),
    )


def _register_describe_script(mcp_server: ExasolMCPServer) -> None:
    mcp_server.tool(
        mcp_server.describe_script,
        description=(
            "The tool describes the specified user defined function (UDF) in "
            "the specified schema of the Exasol Database. It provides the "
            "list of input parameters, the list of emitted parameters or the "
            "SQL type of a single returned value. For each parameter it "
            "provides the name and the SQL type. Both the input and the "
            "emitted parameters can be dynamic or, in other words, flexible. "
            "The dynamic parameters are indicated with ... (triple dot) string "
            "instead of the parameter list. The description includes some usage "
            "notes and a call example."
        ),
    )


def _register_execute_query(mcp_server: ExasolMCPServer) -> None:
    mcp_server.tool(
        mcp_server.execute_query,
        description=(
            "The tool executes the specified query in the Exasol Database. The "
            "query must be a SELECT statement. The tool returns data selected "
            "by the query."
        ),
    )


def register_tools(mcp_server: ExasolMCPServer, config: McpServerSettings) -> None:
    if config.schemas.enable:
        _register_list_schemas(mcp_server)
        _register_find_schemas(mcp_server)
    if config.tables.enable or config.views.enable:
        _register_list_tables(mcp_server)
        _register_find_tables(mcp_server)
    if config.functions.enable:
        _register_list_functions(mcp_server)
        _register_find_functions(mcp_server)
    if config.scripts.enable:
        _register_list_scripts(mcp_server)
        _register_find_scripts(mcp_server)
    if config.columns.enable:
        _register_describe_table(mcp_server)
    if config.parameters.enable:
        _register_describe_function(mcp_server)
        _register_describe_script(mcp_server)
    if config.enable_read_query:
        _register_execute_query(mcp_server)


def get_mcp_settings() -> McpServerSettings:
    """
    Reads optional settings. They can be provided either in a json string stored in the
    EXA_MCP_SETTINGS environment variable or in a json file. In the latter case
    EXA_MCP_SETTINGS must contain the file path.
    """
    try:
        settings_text = os.environ.get(ENV_SETTINGS)
        if not settings_text:
            return McpServerSettings()
        elif re.match(r"^\s*\{.*\}\s*$", settings_text):
            return McpServerSettings.model_validate_json(settings_text)
        elif os.path.isfile(settings_text):
            with open(settings_text) as f:
                return McpServerSettings.model_validate(json.load(f))
        raise ValueError(
            "Invalid MCP Server configuration settings. The configuration "
            "environment variable should either contain a json string or "
            "point to an existing json file."
        )
    except (ValidationError, json.decoder.JSONDecodeError) as config_error:
        raise ValueError("Invalid MCP Server configuration settings.") from config_error


def create_mcp_server(
    connection: DbConnection, config: McpServerSettings
) -> ExasolMCPServer:
    """
    Creates the Exasol MCP Server and registers its tools.
    """
    mcp_server = ExasolMCPServer(connection=connection, config=config)
    register_tools(mcp_server, config)
    return mcp_server


def main():
    """
    Main entry point that creates and runs the MCP server.
    """
    dsn = os.environ[ENV_DSN]
    user = os.environ[ENV_USER]
    password = os.environ[ENV_PASSWORD]
    mcp_settings = get_mcp_settings()

    def connection_factory() -> ExaConnection:
        return pyexasol.connect(
            dsn=dsn, user=user, password=password, fetch_dict=True, compression=True
        )

    connection = DbConnection(connection_factory=connection_factory)

    mcp_server = create_mcp_server(connection=connection, config=mcp_settings)
    mcp_server.run()


if __name__ == "__main__":
    main()
